"""Lambda generator."""

import hashlib
import logging
import os
import typing

import black

import picofun.config
import picofun.template

logger = logging.getLogger(__name__)

LAMBDA_SUFFIX_LENGTH = 4


class LambdaGenerator:
    """Lambda generator."""

    def __init__(
        self,
        template: picofun.template.Template,
        namespace: str,
        config: picofun.config.Config,
    ) -> None:
        """Initialize the lambda generator."""
        self._template = template
        self._config = config

        self.max_length = 64 - len(f"{namespace}_")
        # Remove one for the underscore between the prefix and the suffix.
        self.prefix_length = self.max_length - LAMBDA_SUFFIX_LENGTH - 1

    def _get_name(self, method: str, path: str) -> str:
        clean_path = path.replace("{", "").replace("}", "")
        lambda_name = (
            f"{method}_{clean_path.replace('/', '_').replace('.', '_').strip('_')}"
        )

        if len(lambda_name) > self.max_length:
            suffix = hashlib.sha512(lambda_name.encode()).hexdigest()[
                :LAMBDA_SUFFIX_LENGTH
            ]
            # The underscire adds another character to the length.
            lambda_name = f"{lambda_name[:self.prefix_length]}_{suffix}"

        return lambda_name

    def generate(
        self,
        api_data: dict[str : typing.Any],
    ) -> list[str]:
        """Generate the lambda functions."""
        output_dir = self._config.output_dir

        lambda_dir = os.path.join(output_dir, "lambdas")
        if not os.path.exists(lambda_dir):
            os.makedirs(lambda_dir, exist_ok=True)

        base_url = api_data["servers"][0]["url"]

        lambdas = []
        for path, path_details in api_data["paths"].items():
            for method, details in path_details.items():
                if method not in ["get", "put", "post", "delete", "patch", "head"]:
                    continue

                lambda_name = self._get_name(method, path)

                code = self.render(
                    base_url,
                    method,
                    path,
                    details,
                )

                script_filename = f"{lambda_name}.py"
                script_path = os.path.join(lambda_dir, script_filename)
                with open(script_path, "w") as file:
                    file.write(code)

                logger.info("Generated function: %s", script_path)
                lambdas.append(lambda_name)

        lambdas.sort()
        return lambdas

    def render(
        self, base_url: str, method: str, path: str, details: dict[str : typing.Any]
    ) -> str:
        """Render the lambda function."""
        preprocessor_handler = self._config.preprocessor
        preprocessor = (
            ".".join(preprocessor_handler.split(".")[:-1])
            if preprocessor_handler
            else None
        )

        postprocessor_handler = self._config.postprocessor
        postprocessor = (
            ".".join(postprocessor_handler.split(".")[:-1])
            if postprocessor_handler
            else None
        )

        code = self._template.render(
            "lambda.py.j2",
            base_url=base_url,
            method=method,
            path=path,
            details=details,
            preprocessor=preprocessor,
            preprocessor_handler=preprocessor_handler,
            postprocessor=postprocessor,
            postprocessor_handler=postprocessor_handler,
        )
        return black.format_str(code, mode=black.Mode())



"""Test the requirements manager."""

import os
import tempfile

import picofun.requirements_manager


def test_load() -> None:
    """Test loading the requirements file."""
    requirements = picofun.requirements_manager.RequirementsManager(
        "tests/data/requirements.txt"
    )

    assert requirements._requirements == ["picorun==0.1.0", ""]


def test_load_no_file() -> None:
    """Test loading the requirements file."""
    requirements = picofun.requirements_manager.RequirementsManager(
        "tests/data/requirements-not-found.txt"
    )

    assert requirements._requirements == []


def test_update() -> None:
    """Test updating the requirements file."""
    requirements = picofun.requirements_manager.RequirementsManager(
        "tests/data/requirements.txt"
    )
    requirements.update("picorun", "1.0.0")

    assert requirements._requirements == ["picorun==0.1.0", ""]


def test_update_add() -> None:
    """Test updating the requirements file."""
    requirements = picofun.requirements_manager.RequirementsManager(
        "tests/data/requirements-not-found.txt"
    )
    requirements.update("picorun", "1.0.0")

    assert requirements._requirements == ["picorun==1.0.0"]


def test_save() -> None:
    """Test saving the requirements file."""
    (_, filename) = tempfile.mkstemp(".txt", "picofun_requirements")

    requirements = picofun.requirements_manager.RequirementsManager(filename)
    requirements.update("picorun", "0.1.0")
    requirements.save()

    with open(filename) as file:
        contents = file.readlines()

    assert contents == ["picorun==0.1.0"]

    os.remove(filename)



"""Terraform generator."""

import logging
import os

import picofun.config
import picofun.template

logger = logging.getLogger(__name__)


class TerraformGenerator:
    """Terraform generator."""

    def __init__(
        self,
        template: picofun.template.Template,
        namespace: str,
        config: picofun.config.Config,
    ) -> None:
        """Initialize the terraform generator."""
        self._template = template
        self._namespace = namespace
        self._config = config

    def generate(
        self,
        lambdas: list[str],
    ) -> None:
        """Generate terraform configuration for the Lambda functions."""
        output_dir = self._config.output_dir

        template = self._template.get("main.tf.j2")

        terraform_content = template.render(
            bundle=self._config.bundle,
            iam_role_prefix=self._config.iam_role_prefix,
            lambdas=lambdas,
            layers=self._config.layers,
            namespace=self._namespace,
            role_permissions_boundary=self._config.role_permissions_boundary,
            subnets=self._config.subnets,
            tags=self._config.tags,
        )
        output_file = os.path.join(output_dir, "main.tf")
        with open(output_file, "w") as terraform_file:
            terraform_file.write(terraform_content)

        logger.info("Generated terraform: %s", output_file)



"""Test picofun Configuration loader."""

import os
import random
import string
import tempfile

import pytest

import picofun.config


def test_config_load() -> None:
    """Test loading a configuration file."""
    config = picofun.config.Config("tests/data/config.toml")

    assert config.output_dir == os.path.realpath("tests/data/output")
    assert config.layers == ["arn:aws:lambda:us-east-1:012345678910:layer:example:1"]
    assert config.template_path == "tests/data/templates"


def test_config_load_empty() -> None:
    """Test loading an empty configuration file."""
    config = picofun.config.Config("tests/data/empty.toml")

    assert config.output_dir == os.path.realpath(f"{os.getcwd()}/output")
    assert config.layers == []


def test_config_load_missing() -> None:
    """Test loading a missing configuration file."""
    with pytest.raises(picofun.errors.ConfigFileNotFoundError):
        picofun.config.Config("tests/data/missing.toml")


def test_config_load_invalid() -> None:
    """Test loading an invalid configuration file."""
    with pytest.raises(picofun.errors.InvalidConfigError):
        picofun.config.Config("tests/data/invalid.toml")


def test_config_load_invalid_value() -> None:
    """Test loading an invalid configuration file."""
    with pytest.raises(picofun.errors.UnknownConfigValueError):
        picofun.config.Config("tests/data/invalid_type.toml")


def test_config_load_relative_path() -> None:
    """Test loading a configuration file with a relative path."""
    config = picofun.config.Config("tests/data/config_relative.toml")

    assert config.output_dir == os.path.realpath("tests/data/output")


def test_config_load_relative_path_missing() -> None:
    """Test loading a configuration file with a relative path."""
    with pytest.raises(picofun.errors.ConfigFileNotFoundError):
        picofun.config.Config("tests/data/missing.toml")


def test_config_load_absolute_path_missing() -> None:
    """Test loading a configuration file with an absolute path."""
    with pytest.raises(picofun.errors.ConfigFileNotFoundError):
        picofun.config.Config(
            "/tmp/invalid/config/file/location/45c5391e-6e9a-496e-8694-9ef1a03addb3.toml"
        )


def test_config_getattr() -> None:
    """Test loading a configuration file."""
    config = picofun.config.Config("tests/data/config.toml")

    assert config.layers == ["arn:aws:lambda:us-east-1:012345678910:layer:example:1"]


def test_config_getattr_invalid() -> None:
    """Test __getattr__ with an invalid property."""
    config = picofun.config.Config("tests/data/config.toml")
    with pytest.raises(picofun.errors.UnknownConfigValueError):
        _ = config.invalid


def test_config_setattr_invalid() -> None:
    """Test splitting layers string into a list."""
    config = picofun.config.Config("tests/data/empty.toml")

    with pytest.raises(picofun.errors.UnknownConfigValueError):
        config.invalid = "invalid"


def test_config_setattr_layers_empty() -> None:
    """Test splitting layers string into a list."""
    config = picofun.config.Config("tests/data/empty.toml")

    config.layers = ""
    assert config.layers == []


def test_config_setattr_layers() -> None:
    """Test splitting layers string into a list."""
    layers = [
        "arn:aws:lambda:us-east-1:012345678910:layer:example:1",
        "arn:aws:lambda:us-east-1:012345678910:layer:anotherexample:123",
    ]

    config = picofun.config.Config("tests/data/empty.toml")
    assert config.layers == []

    config.layers = ", ".join(layers)
    assert config.layers == layers


def test_config_load_absolute_path() -> None:
    """Test loading a configuration file with an absolute path."""
    config = picofun.config.Config(os.path.realpath("tests/data/config.toml"))

    assert config.output_dir == os.path.realpath("tests/data/output")


def test_fix_target_path_relative() -> None:
    """Test fixing the target path with a realtive path."""
    config = picofun.config.Config("tests/data/config.toml")

    assert config._fix_target_path("tests/data") == os.path.realpath("tests/data")


def test_fix_target_path_absolute() -> None:
    """Test fixing the target path with an absolute path."""
    config = picofun.config.Config("tests/data/config.toml")

    assert config._fix_target_path("/tmp") == os.path.realpath("/tmp")


def test_fix_target_path_dots() -> None:
    """Test fixing the target path containing dots."""
    config = picofun.config.Config("tests/data/config.toml")

    assert config._fix_target_path("tests/../tests/data") == os.path.realpath(
        "tests/data"
    )


def test_fix_target_path_empty() -> None:
    """Test fixing the target path with an empty path."""
    config = picofun.config.Config("tests/data/config.toml")

    assert config._fix_target_path("") == os.path.realpath(os.getcwd())


def test_fix_target_path_none() -> None:
    """Test fixing the target path with an empty path."""
    config = picofun.config.Config("tests/data/config.toml")

    assert config._fix_target_path() == os.path.realpath(os.getcwd())


def test_fix_target_path_missing() -> None:
    """Test fixing the target path with an empty path."""
    config = picofun.config.Config("tests/data/config.toml")

    assert config._fix_target_path("missing") == os.path.realpath(
        os.path.join(os.getcwd(), "missing")
    )


def test_fix_target_path_missing_absolute() -> None:
    """Test fixing the target path with an empty path."""
    config = picofun.config.Config("tests/data/config.toml")

    assert config._fix_target_path("/missing") == os.path.realpath("/missing")


def test_fix_target_path_missing_relative() -> None:
    """Test fixing the target path with an empty path."""
    config = picofun.config.Config("tests/data/config.toml")

    assert config._fix_target_path("missing") == os.path.realpath(
        os.path.join(os.getcwd(), "missing")
    )


def test_fix_bundle_path_relative() -> None:
    """Test fixing the bundle path with a realtive path."""
    config = picofun.config.Config("tests/data/config.toml")

    assert config._fix_bundle_path("subdir") == os.path.realpath("tests/data/subdir")


def test_fix_bundle_path_absolute() -> None:
    """Test fixing the bundle path with an absolute path."""
    config = picofun.config.Config("tests/data/config.toml")

    assert config._fix_bundle_path("/tmp") == os.path.realpath("/tmp")


def test_fix_bundle_path_dots() -> None:
    """Test fixing the bundle path containing dots."""
    config = picofun.config.Config("tests/data/config.toml")

    assert config._fix_bundle_path("../data") == os.path.realpath("tests/data")


def test_fix_bundle_path_empty() -> None:
    """Test fixing the bundle path with an empty path."""
    config = picofun.config.Config("tests/data/config.toml")

    assert config._fix_bundle_path("") == os.path.realpath("tests/data")


def test_load_from_file() -> None:
    """Test loading a configuration file."""
    config = picofun.config.Config("tests/data/config.toml")

    with tempfile.NamedTemporaryFile() as tmp:
        tmp.write(b"output_dir = 'tests/data/output'")
        tmp.seek(0)

        config.load_from_file(tmp.name)

        assert config.output_dir == os.path.realpath("tests/data/output")


def test_load_from_file_missing() -> None:
    """Test loading a configuration file."""
    config = picofun.config.Config("tests/data/config.toml")

    with pytest.raises(picofun.errors.ConfigFileNotFoundError):
        config.load_from_file(
            "/tmp/invalid/config/file/location/45c5391e-6e9a-496e-8694-9ef1a03addb3.toml"
        )


def test_load_from_file_invalid() -> None:
    """Test loading a configuration file."""
    config = picofun.config.Config("tests/data/config.toml")

    with tempfile.NamedTemporaryFile() as tmp:
        tmp.write(b"this = invalid = toml")
        tmp.seek(0)

        with pytest.raises(picofun.errors.InvalidConfigError):
            config.load_from_file(tmp.name)


def test_merge() -> None:
    """Test merging two configurations."""
    config = picofun.config.Config("tests/data/config.toml")

    config.merge(output_dir="tests/output")

    assert config.output_dir == os.path.realpath("tests/output")


def test_merge_invalid() -> None:
    """Test merging two configurations."""
    config = picofun.config.Config("tests/data/config.toml")

    with pytest.raises(picofun.errors.UnknownConfigValueError):
        config.merge(invalid="invalid")


def test_merge_invalid_type() -> None:
    """Test merging two configurations."""
    config = picofun.config.Config("tests/data/config.toml")

    with pytest.raises(picofun.errors.InvalidConfigTypeError):
        config.merge(output_dir=1234)


def test_set_defaults() -> None:
    """Test setting defaults."""
    config = picofun.config.Config("tests/data/empty.toml")

    config.set_defaults()

    assert config.bundle is None
    assert config.layers == []


def test_asdict() -> None:
    """Test converting the config to a dictionary."""
    config = picofun.config.Config("tests/data/config.toml")

    config_dict = config.asdict()

    assert isinstance(config_dict, dict)
    assert config_dict["output_dir"] == os.path.realpath("tests/data/output")


def test_asdict_empty() -> None:
    """Test converting the config to a dictionary."""
    config = picofun.config.Config("tests/data/empty.toml")

    assert config.asdict()["layers"] == []


def test_get_config_file() -> None:
    """Test getting the path to the config file."""
    path = os.path.join(os.path.dirname(__file__), "data")
    config = picofun.config.Config(path)
    assert config._config_file == f"{path}/picofun.toml"


def test_get_config_file_empty() -> None:
    """Test getting the path to the config file with an empty argument."""
    path = os.getcwd()
    config = picofun.config.Config()
    assert config._config_file == f"{path}/picofun.toml"


def test_get_config_file_invalid() -> None:
    """Test not getting the config file from an invalid path."""
    with pytest.raises(picofun.errors.ConfigFileNotFoundError):
        picofun.config.Config("/invalid/")


def test_setattr_output_dir_mkdir() -> None:
    """Test creating the output directory."""
    config = picofun.config.Config("tests/data/empty.toml")

    tmp = os.path.realpath(
        f'/tmp/picofun_tests_{"".join(random.sample(string.ascii_letters + string.digits, 16))}'
    )

    assert not os.path.exists(tmp)
    config.output_dir = tmp
    assert os.path.exists(tmp)

    os.rmdir(tmp)



"""Template loader for Jinja2 templates."""

import os
import typing

import jinja2


class Template:
    """Manages Jinja2 templates."""

    def __init__(self, base_path: str | None = "") -> None:
        """Initialise TemplateLoader."""
        if not os.path.isabs(base_path):
            base_path = os.path.join(os.getcwd(), base_path)

        path = os.path.realpath(base_path)
        loader = jinja2.FileSystemLoader(path)
        self._environment = jinja2.Environment(loader=loader)  # noqa: S701 We're not generating HTML # nosec

        self.templates = {}

    def get(self, filename: str) -> jinja2.Template:
        """Get a template by filename."""
        if filename not in self.templates:
            try:
                self.templates[filename] = self._environment.get_template(filename)
            except jinja2.exceptions.TemplateNotFound as e:
                raise FileNotFoundError(e.name) from e

        return self.templates[filename]

    def render(self, filename: str, **kwargs: dict[str : typing.Any]) -> str:
        """Render a template by filename."""
        return self.get(filename).render(**kwargs)



"""Tests for picofun exceptions."""

import pytest
import requests.exceptions
import requests.models

import picofun.errors


def test_config_file_not_found_error() -> None:
    """Test ConfigFileError."""
    with pytest.raises(picofun.errors.ConfigFileNotFoundError):
        raise picofun.errors.ConfigFileNotFoundError("missing.toml")


def test_download_spec_error() -> None:
    """Test DownloadSpecError."""
    exception = requests.exceptions.RequestException()
    with pytest.raises(picofun.errors.DownloadSpecError):
        raise picofun.errors.DownloadSpecError(exception)


def test_download_spec_http_error() -> None:
    """Test DownloadSpecHTTPError."""
    response = requests.models.Response()
    response.status_code = 404
    response.reason = "Not Found"

    exception = requests.exceptions.HTTPError(response=response)

    with pytest.raises(picofun.errors.DownloadSpecHTTPError):
        raise picofun.errors.DownloadSpecHTTPError(exception)


def test_invalid_config_error() -> None:
    """Test InvalidConfigError."""
    with pytest.raises(picofun.errors.InvalidConfigError):
        raise picofun.errors.InvalidConfigError()


def test_invalid_spec_error() -> None:
    """Test InvalidSpecError."""
    with pytest.raises(picofun.errors.InvalidSpecError):
        raise picofun.errors.InvalidSpecError()


def test_unknown_config_value_error() -> None:
    """Test UnknownConfigValueError."""
    with pytest.raises(picofun.errors.UnknownConfigValueError):
        raise picofun.errors.UnknownConfigValueError("missing")


def test_invalid_config_type_error() -> None:
    """Test InvalidConfigTypeError."""
    with pytest.raises(picofun.errors.InvalidConfigTypeError):
        raise picofun.errors.InvalidConfigTypeError("example", [], str)



"""Spec file handler."""

import abc
import errno
import json
import os

import requests
import yaml

import picofun.errors


class SpecLoader(abc.ABC):  # pragma: no cover
    """Loads an OpenAPI spec file."""

    @abc.abstractmethod
    def load(self, location: str) -> str:
        """
        Load a spec file.

        :param location: The location of the file.
        :return: The contents of the file.
        """
        pass


class FileSpecLoader(SpecLoader):
    """Loads an OpenAPI spec file from disk."""

    def load(self, location: str) -> str:
        """
        Load a spec file from disk.

        :param location: The location of the file.
        :raises FileNotFoundError: If the file does not exist.
        :return: The contents of the file.
        """
        if not os.path.exists(location):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), location)

        with open(location, "rb") as file:
            return file.read()


class HTTPSpecLoader(SpecLoader):
    """Loads an OpenAPI spec file from a URL."""

    def load(self, location: str) -> str:
        """
        Load a spec file from a URL.

        :param location: The location of the file.
        :raises DownloadSpecError: If the file cannot be downloaded.
        :return: The contents of the file.
        """
        try:
            response = requests.get(location, timeout=30)
            response.raise_for_status()
        except requests.HTTPError as e:
            raise picofun.errors.DownloadSpecError(e) from e
        except requests.RequestException as e:
            raise picofun.errors.DownloadSpecError(e) from e

        return response.content


class SpecParser(abc.ABC):  # pragma: no cover
    """Parses an OpenAPI spec file."""

    @abc.abstractmethod
    def parse(self, content: str) -> dict:
        """
        Parse the contents of the spec file.

        :param content: The raw string containing the spec file.
        :raises InvalidSpecError: If the spec file is not valid.
        :return: The contents of the spec file as a dict.
        """
        pass


class JSONSpecParser(SpecParser):
    """Parses an OpenAPI spec file in JSON format."""

    def parse(self, content: str) -> dict:
        """
        Parse the contents of the spec file.

        :param content: The raw string containing the spec file.
        :raises InvalidSpecError: If the spec file is not valid JSON.
        :return: The contents of the spec file as a dict.
        """
        try:
            return json.loads(content)
        except ValueError as e:
            raise picofun.errors.InvalidSpecError() from e


class YAMLSpecParser(SpecParser):
    """Parses an OpenAPI spec file in YAML format."""

    def patch_loader(self, loader: yaml.SafeLoader) -> yaml.SafeLoader:
        """
        Patch the loader to allow = as a value.

        Handles being call multiple times. It only patches the loader on the first call.

        See https://github.com/yaml/pyyaml/issues/89 for more info.

        :return: The patched YAML loader.
        """
        if "=" in loader.yaml_implicit_resolvers:
            loader.yaml_implicit_resolvers.pop("=")
        return loader

    def parse(self, content: str) -> dict:
        """
        Parse the contents of the spec file.

        :param content: The raw string containing the spec file.
        :raises InvalidSpecError: If the spec file is not valid YAML.
        :return: The contents of the spec file as a dict.
        """
        loader = yaml.SafeLoader
        patched_loader = self.patch_loader(loader)
        try:
            return yaml.load(
                content,
                Loader=patched_loader,  # noqa: S506 Uses a patched safe loader
            )
        except yaml.YAMLError as e:
            raise picofun.errors.InvalidSpecError() from e


class Spec:
    """Spec file handler."""

    def __init__(self, location: str) -> None:
        """
        Initialise Spec.

        :param source: The location of the spec file. Can be a URL or a path to a file.
        """
        self._content = self._load(location)

    def _load(self, location: str) -> str:
        """
        Load a spec file.

        :param location: The location of the spec file. Can be a URL or a path to a file.
        :return: The contents of the spec file as a string.
        """
        loader = HTTPSpecLoader() if location.startswith("http") else FileSpecLoader()

        return loader.load(location)

    def parse(self) -> dict:
        """
        Parse the contents of the spec file.

        :param content: The raw string containing the spec file
        :raises InvalidSpecError: If the spec file is not valid JSON or YAML
        :return: The contents of the spec file as a dict
        """
        try:
            return JSONSpecParser().parse(self._content)
        except picofun.errors.InvalidSpecError:
            # Try YAML parser if JSON parser fails
            pass

        try:
            return YAMLSpecParser().parse(self._content)
        except picofun.errors.InvalidSpecError as e:
            raise picofun.errors.InvalidSpecError() from e



"""Command line interface for picofun."""

import typing

import typer

import picofun.config
import picofun.lambda_generator
import picofun.layer
import picofun.spec
import picofun.template
import picofun.terraform_generator

app = typer.Typer()


@app.command()
def main(
    namespace: typing.Annotated[
        str, typer.Argument(help="Namespace for the generated functions")
    ],
    spec_file: typing.Annotated[
        str, typer.Argument(help="URL or path to the OpenAPI spec file")
    ],
    config_file: typing.Annotated[
        typing.Optional[str], typer.Option(help="Full path to the configuration file")
    ] = None,
    output_dir: typing.Annotated[
        typing.Optional[str],
        typer.Option(help="Directory to output the generated files"),
    ] = None,
    layers: typing.Annotated[
        typing.Optional[str],
        typer.Option(help="Comma separated list of Lambda layer ARNs"),
    ] = "",
    bundle: typing.Annotated[
        typing.Optional[str],
        typer.Option(
            help="Path to code to bundle into a layer. If requirements.txt present pip install will be run."
        ),
    ] = None,
) -> None:
    """Generate lambda functions and terraform configuration to call REST APIs."""
    config = picofun.config.Config(config_file)
    config.merge(
        output_dir=output_dir,
        layers=layers,
        bundle=bundle,
    )

    spec = picofun.spec.Spec(spec_file)
    api_data = spec.parse()

    template = picofun.template.Template(config.template_path)

    lambda_generator = picofun.lambda_generator.LambdaGenerator(
        template, namespace, config
    )
    lambdas = lambda_generator.generate(api_data)

    layer = picofun.layer.Layer(config)
    layer.prepare()

    terraform_generator = picofun.terraform_generator.TerraformGenerator(
        template, namespace, config
    )
    terraform_generator.generate(lambdas)



"""Lambda layer manager."""

import logging
import os
import shutil

import picofun.requirements_manager

logger = logging.getLogger(__name__)

PICORUN_VERSION = "0.2.1"


class Layer:
    """Manage the contents of a Lambda layer."""

    def __init__(self, config: dict) -> None:
        """Initialise the layer."""
        self._config = config

    def prepare(self) -> None:
        """Prepare the contents of a Lambda layer."""
        output_dir = self._config.output_dir
        layer_dir = os.path.join(output_dir, "layer")

        if self._config.bundle and os.path.isdir(self._config.bundle):
            shutil.copytree(self._config.bundle, layer_dir)

        # Handle requirements.txt updates
        requirements_file = os.path.join(layer_dir, "requirements.txt")
        requirements = picofun.requirements_manager.RequirementsManager(
            requirements_file
        )
        requirements.update("picorun", PICORUN_VERSION)
        requirements.save()

        logger.info("Prepared layer contents: %s", layer_dir)



"""Configuration handling."""

import os
import typing
from importlib.resources import files

import tomli as toml

import picofun.errors


class Config:
    """Configuration management class."""

    _attrs: typing.ClassVar[dict[str : typing.Any]] = {
        "_config_file": str,
        "bundle": str,
        "iam_role_prefix": str,
        "layers": list,
        "output_dir": str,
        "postprocessor": str,
        "preprocessor": str,
        "role_permissions_boundary": str,
        "subnets": list,
        "tags": dict,
        "template_path": str,
    }

    def __init__(self, file_path: typing.Optional[str] = None) -> None:
        """
        Initialize the configuration.

        :param file_path: The path to the configuration file.
        """
        self.set_defaults()
        self._config_file = self._get_config_file(file_path)
        self.load_from_file(self._config_file)

    def _get_config_file(self, config_file: typing.Optional[str] = None) -> str:
        """Get the path to the configuration file."""
        if not config_file:
            config_file = os.getcwd()

        if os.path.isdir(config_file):
            config_file = os.path.join(config_file, "picofun.toml")

        if os.path.isfile(config_file):
            return config_file

        raise picofun.errors.ConfigFileNotFoundError(config_file)

    def __getattr__(self, name: str) -> str | list | dict[str, typing.Any]:
        """
        Get a configuration value.

        :param name: The name of the configuration value.
        :return: The configuration value.
        """
        if name not in self._attrs:
            raise picofun.errors.UnknownConfigValueError(name)

    def __setattr__(self, name: str, value: str | list | dict[str, typing.Any]) -> None:
        """
        Set a configuration value.

        :param name: The name of the configuration value.
        :param value: The value to set.
        """
        if name not in self._attrs:
            raise picofun.errors.UnknownConfigValueError(name)

        # First allow layers to be set as a comma separated string and convert it to a list
        if name == "layers" and isinstance(value, str):
            value = [raw.strip() for raw in value.split(",")]
            if value == [""]:
                value = []

        if not isinstance(value, self._attrs[name]):
            raise picofun.errors.InvalidConfigTypeError(name, value, self._attrs[name])

        if name == "bundle" and value:
            value = self._fix_bundle_path(value)

        if name == "output_dir":
            value = self._fix_target_path(value)
            # Ensure the path exists
            if not os.path.exists(value):
                os.makedirs(value, exist_ok=True)

        self.__dict__[name] = value

    def _fix_bundle_path(self, bundle_path: str) -> str:
        """
        Fix the bundle path to ensure it is absolute.

        :param bundle_path: The path to the bundle.
        :return: The absolute path to the bundle.
        """
        if not os.path.isabs(bundle_path):
            bundle_path = os.path.join(os.path.dirname(self._config_file), bundle_path)

        return os.path.realpath(bundle_path)

    def _fix_target_path(self, output_dir: str = "") -> str:
        """
        Fix the target path to ensure it is absolute.

        :param output_dir: The name of the output directory
        :return: The absolute path to the output directory
        """
        if os.path.isabs(output_dir):
            return os.path.realpath(output_dir)

        return os.path.realpath(os.path.join(os.getcwd(), output_dir))

    def load_from_file(self, path: str) -> None:
        """
        Load configuration from toml file.

        :param path: The path to the configuration file.
        :return: The configuration.
        """
        if not os.path.isfile(path):
            raise picofun.errors.ConfigFileNotFoundError(path=path)

        with open(path, "rb") as file:
            try:
                file_config = toml.load(file)
            except toml.TOMLDecodeError as e:
                raise picofun.errors.InvalidConfigError() from e

        self.merge(**file_config)

    def merge(self, **kwargs: dict[str, typing.Any]) -> None:
        """
        Merge the configuration with the CLI arguments.

        :param kwargs: The CLI arguments.
        """
        for key in kwargs:
            if key not in self._attrs:
                raise picofun.errors.UnknownConfigValueError(key)

            if kwargs[key] is not None:
                setattr(self, key, kwargs[key])

    def set_defaults(self) -> None:
        """Set the default values for the configuration."""
        defaults = {
            "_config_file": "",
            "bundle": None,
            "iam_role_prefix": "pf-",
            "layers": [],
            "output_dir": os.path.realpath(os.path.join(os.getcwd(), "output")),
            "postprocessor": "",
            "preprocessor": "",
            "role_permissions_boundary": None,
            "subnets": [],
            "tags": {},
            "template_path": os.path.join(files("picofun"), "templates"),
        }

        for key in defaults:
            self.__dict__[key] = defaults[key]

    def asdict(self) -> dict[str, typing.Any]:
        """
        Convert the configuration to a dictionary.

        :return: The configuration dictionary.
        """
        return self.__dict__



"""Tests for the TerraformGenerator class."""

import tempfile

import picofun.config
import picofun.template
import picofun.terraform_generator


def test_generate() -> None:
    """Test generating the terraform configuration."""
    config = picofun.config.Config("tests/data/empty.toml")
    tpl = picofun.template.Template("tests/data/templates")
    with tempfile.TemporaryDirectory() as out_dir:
        config.output_dir = out_dir
        generator = picofun.terraform_generator.TerraformGenerator(tpl, "test", config)
        generator.generate([])



"""Tests for the LambdaGenerator class."""

import tempfile

import picofun.config
import picofun.lambda_generator
import picofun.template


def test_get_name() -> None:
    """Test getting the name of the lambda."""
    tpl = picofun.template.Template("tests/data/templates")
    generator = picofun.lambda_generator.LambdaGenerator(tpl, "", "tests/data/lambda")

    assert generator._get_name("method", "path") == "method_path"


def test_get_name_with_dot() -> None:
    """Test getting the name of the lambda when path contains a dot (.)."""
    tpl = picofun.template.Template("tests/data/templates")
    generator = picofun.lambda_generator.LambdaGenerator(tpl, "", "tests/data/lambda")

    assert generator._get_name("method", "path.json") == "method_path_json"


def test_get_name_too_long() -> None:
    """Test getting the name of the lambda."""
    tpl = picofun.template.Template("tests/data/templates")
    generator = picofun.lambda_generator.LambdaGenerator(
        tpl, "prefix", "tests/data/lambda"
    )

    path = "how_long_does_this_string_need_to_be_before_it_is_truncated_by_get_name"

    name = generator._get_name("get", path)
    assert len(name) == 57  # 64 - "prefix_" = 57
    assert name[:52] == "get_how_long_does_this_string_need_to_be_before_it_i"


def test_generate() -> None:
    """Test generating the lambda functions."""
    spec = {
        "swagger": "2.0",
        "info": {"title": "Simple API overview", "version": "v2"},
        "servers": [{"url": "https://example.com"}],
        "paths": {
            "/": {
                "get": {
                    "operationId": "example",
                    "summary": "Example endpoint",
                    "produces": ["application/json"],
                    "responses": {
                        "200": {
                            "description": "OK response",
                        },
                        "404": {
                            "description": "Not found response",
                        },
                    },
                }
            },
        },
    }

    tpl = picofun.template.Template("tests/data/templates")
    config = picofun.config.Config("tests/data/empty.toml")
    with tempfile.TemporaryDirectory() as out_dir:
        config.output_dir = out_dir
        generator = picofun.lambda_generator.LambdaGenerator(tpl, "", config)

        assert generator.generate(spec) == ["get_"]


def test_generate_empty() -> None:
    """Test generate with no lambdas."""
    spec = {
        "swagger": "2.0",
        "info": {"title": "Simple API overview", "version": "v2"},
        "servers": [{"url": "https://example.com"}],
        "paths": {
            "/": {
                "invalid": {
                    "operationId": "example",
                    "summary": "Example endpoint",
                    "produces": ["application/json"],
                    "responses": {
                        "200": {
                            "description": "OK response",
                        },
                        "404": {
                            "description": "Not found response",
                        },
                    },
                }
            },
        },
    }

    tpl = picofun.template.Template("tests/data/templates")
    config = picofun.config.Config("tests/data/empty.toml")

    with tempfile.TemporaryDirectory() as out_dir:
        config.output_dir = out_dir
        generator = picofun.lambda_generator.LambdaGenerator(tpl, "", config)
        assert generator.generate(spec) == []


def test_generate_invalid_method() -> None:
    """Test generate with invalid HTTP method."""
    spec = {"servers": [{"url": "https://example.com"}], "paths": {}}

    tpl = picofun.template.Template("tests/data/templates")
    config = picofun.config.Config("tests/data/empty.toml")

    with tempfile.TemporaryDirectory() as out_dir:
        config.output_dir = out_dir
        generator = picofun.lambda_generator.LambdaGenerator(tpl, "", config)
        assert generator.generate(spec) == []


def test_render() -> None:
    """Test rendering the lambda."""
    tpl = picofun.template.Template("tests/data/templates")
    config = picofun.config.Config("tests/data/empty.toml")
    generator = picofun.lambda_generator.LambdaGenerator(tpl, "", config)

    code = generator.render("https://example.com", "get", "/path", {})

    assert (
        code
        == """\"\"\"Generated AWS Lambda function for making an API call.\"\"\"

from typing import Any

import requests
from aws_lambda_powertools import Logger
from aws_lambda_powertools.utilities.typing import LambdaContext

import picorun


logger = Logger()


@logger.inject_lambda_context
def handler(event: dict[str, Any], context: LambdaContext) -> dict[str, Any]:
    \"\"\".\"\"\"
    properties = picorun.ApiRequestArgs(**event)

    url = "https://example.com/path".format(**properties.path)

    timeout = context.get_remaining_time_in_millis() - 1000
    response = requests.get(url, timeout=timeout, **properties.to_kwargs())
    out = picorun.ApiResponse(response)

    if response.status_code >= 400:
        logger.error("Error %d: %s", response.status_code, response.text)
        raise picorun.ApiError(response.text, response.status_code)

    return out.asdict()
"""
    )



"""Test the spec file handler."""

import pytest
import requests

import picofun.errors
import picofun.spec


def mock_requests_get_spec(*args: list, **kwargs: dict) -> "MockResponse":  # noqa: F821
    """
    Mock requests.get() responses.

    :param args: The positional arguments passed to requests.get().
    :param kwargs: The keyword arguments passed to requests.get().

    :return: A MockResponse object.
    """

    class MockResponse:
        def __init__(self, content: str, status_code: int) -> None:
            """
            Initialise MockResponse.

            :param content: The content of the response.
            :param status_code: The status code of the response.
            """
            self.content = content
            self.status_code = status_code

        def raise_for_status(self) -> None:
            if self.status_code != 200:
                raise requests.HTTPError(self.status_code)

            return

    if args[0] == "https://petstore3.swagger.io/api/v3/openapi.yaml":
        with open("tests/data/petstore.yaml") as file:
            return MockResponse(file.read(), 200)

    if args[0] == "httpinvalid://example.com/":
        raise requests.RequestException()

    return MockResponse("", 404)


def test_spec_parser() -> None:
    """Test the SpecParser interface."""
    with pytest.raises(TypeError):
        picofun.spec.SpecParser()


def test_json_spec_parser() -> None:
    """Test the JSONSpecParser class."""
    parser = picofun.spec.JSONSpecParser()
    assert parser.parse('{"foo": "bar"}') == {"foo": "bar"}


def test_json_spec_parser_invalid_json() -> None:
    """Test the JSONSpecParser class with invalid JSON."""
    parser = picofun.spec.JSONSpecParser()
    with pytest.raises(picofun.errors.InvalidSpecError):
        parser.parse('{"foo": "bar"')


def test_yaml_spec_parser() -> None:
    """Test the YAMLSpecParser class."""
    parser = picofun.spec.YAMLSpecParser()
    assert parser.parse("foo: bar") == {"foo": "bar"}


def test_yaml_spec_parser_invalid_yaml() -> None:
    """Test the YAMLSpecParser class with invalid YAML."""
    parser = picofun.spec.YAMLSpecParser()
    with pytest.raises(picofun.errors.InvalidSpecError):
        parser.parse("foo: b: a: r")


def test_yaml_spec_parser_value_equals() -> None:
    """Test the YAMLSpecParser class with a plus symbol as the value."""
    parser = picofun.spec.YAMLSpecParser()
    assert parser.parse("key: +") == {"key": "+"}


def test_spec_json() -> None:
    """Test the Spec class with JSON."""
    spec = picofun.spec.Spec("tests/data/petstore.json")
    assert spec.parse()["servers"][0]["url"] == "http://petstore.swagger.io/v1"


def test_spec_json_invalid_json() -> None:
    """Test the Spec class with invalid JSON."""
    spec = picofun.spec.Spec("tests/data/broken.json")
    with pytest.raises(picofun.errors.InvalidSpecError):
        spec.parse()


def test_spec_yaml() -> None:
    """Test the Spec class with YAML."""
    spec = picofun.spec.Spec("tests/data/petstore.yaml")
    assert spec.parse()["servers"][0]["url"] == "http://petstore.swagger.io/v1"


def test_spec_yaml_invalid_yaml() -> None:
    """Test the Spec class with invalid YAML."""
    spec = picofun.spec.Spec("tests/data/broken.yaml")
    with pytest.raises(picofun.errors.InvalidSpecError):
        spec.parse()


def test_spec_unknown_file_type() -> None:
    """Test the Spec class with an unknown file type."""
    spec = picofun.spec.Spec("tests/data/lambda.png")
    with pytest.raises(picofun.errors.InvalidSpecError):
        spec.parse()


def test_spec_unknown_file() -> None:
    """Test the Spec class with an unknown file."""
    with pytest.raises(FileNotFoundError):
        picofun.spec.Spec("tests/data/unknown")


def test_spec_http() -> None:
    """Test the Spec class with a URL."""
    import requests

    monkeypatch = pytest.MonkeyPatch()
    monkeypatch.setattr(requests, "get", mock_requests_get_spec)

    spec = picofun.spec.Spec("https://petstore3.swagger.io/api/v3/openapi.yaml")
    assert spec.parse()["servers"][0]["url"] == "http://petstore.swagger.io/v1"


def test_spec_http_missing() -> None:
    """Test the Spec class with a missing URL."""
    import requests

    monkeypatch = pytest.MonkeyPatch()
    monkeypatch.setattr(requests, "get", mock_requests_get_spec)

    with pytest.raises(picofun.errors.DownloadSpecError):
        picofun.spec.Spec("https://example.swagger.io/api/v3/missing.json")


def test_spec_http_invalid() -> None:
    """Test the Spec class with a missing URL."""
    import requests

    monkeypatch = pytest.MonkeyPatch()
    monkeypatch.setattr(requests, "get", mock_requests_get_spec)

    with pytest.raises(picofun.errors.DownloadSpecError):
        picofun.spec.Spec("httpinvalid://example.com/")



"""Tests for the LayerGenerator class."""

import os
import tempfile

import picofun.config
import picofun.layer


def test_prepare() -> None:
    """Test generating the layer."""
    config = picofun.config.Config("tests/data/empty.toml")
    with tempfile.TemporaryDirectory() as out_dir:
        config.output_dir = out_dir
        layer = picofun.layer.Layer(config)
        layer.prepare()

        assert os.path.isfile(os.path.join(out_dir, "layer", "requirements.txt"))


def test_prepare_with_bundle() -> None:
    """Test generating the layer."""
    config = picofun.config.Config("tests/data/empty.toml")

    with tempfile.TemporaryDirectory() as bundle_dir:
        config.bundle = bundle_dir

        with tempfile.TemporaryDirectory() as out_dir:
            (fp, filename) = tempfile.mkstemp(dir=bundle_dir)
            config.output_dir = out_dir

            assert os.path.isfile(filename)
            layer = picofun.layer.Layer(config)
            layer.prepare()

            basename = os.path.basename(filename)
            assert os.path.isfile(os.path.join(out_dir, "layer", basename))



"""PicoFun exceptions."""

import typing

import requests


class ConfigFileNotFoundError(FileNotFoundError):
    """Exception thrown when the config file cannot be found."""

    def __init__(self, path: str) -> None:
        """
        Initialise ConfigFileNotFoundError.

        :param path: The path to the config file.
        """
        super().__init__(f"Config file not found: {path}")


class DownloadSpecError(Exception):
    """Exception thrown when requests is unable to download the spec file."""

    def __init__(self: "DownloadSpecError", arg: requests.RequestException) -> None:
        """
        Initialise DownloadSpecError.

        :param arg: The exception thrown.
        """
        super().__init__(f"Request Error: {arg}")


class DownloadSpecHTTPError(Exception):
    """Exception thrown when requests is unable to download the spec file."""

    def __init__(self, arg: requests.HTTPError) -> None:
        """
        Initialise DownloadSpecHTTPError.

        :param arg: The HTTP error thrown.
        """
        super().__init__(
            f"HTTP Error {arg.response.status_code}: {arg.response.reason}"
        )


class InvalidConfigError(Exception):
    """Exception thrown when the config file is not valid TOML."""

    def __init__(self) -> None:
        """Initialise InvalidConfigError."""
        super().__init__("The config file isn't valid TOML")


class InvalidConfigTypeError(TypeError):
    """Exception thrown when an unknown configuration type is specified."""

    def __init__(
        self,
        name: str,
        value: typing.Any,  # noqa ANN001 We need dynamic types as this generic exception.
        expected: typing.Any,  # noqa ANN401 We need dynamic types as this generic exception.
    ) -> None:
        """Initialise InvalidConfigTypeError."""
        super().__init__(
            f"Expected {expected} for {name}, but receieved {type(value)}."
        )


class InvalidSpecError(Exception):
    """Exception thrown when the spec file is not valid JSON or YAML."""

    def __init__(self) -> None:
        """Initialise InvalidSpecError."""
        super().__init__("The spec file isn't valid JSON or YAML")


class UnknownConfigValueError(AttributeError):
    """Exception thrown when an unknown configuration value is specified."""

    def __init__(self, name: str) -> None:
        """Initialise UnknownConfigValueError."""
        super().__init__(name=name, obj="Config")



"""Test the template module."""

import jinja2
import pytest

import picofun.errors
import picofun.template


def test_template_load() -> None:
    """Test loading a template."""
    template = picofun.template.Template("tests/data/templates")
    file = template.get("empty.j2")
    assert isinstance(file, jinja2.environment.Template)


def test_template_load_missing() -> None:
    """Test loading a missing template."""
    template = picofun.template.Template("tests/data")
    with pytest.raises(FileNotFoundError):
        template.get("missing.j2")


def test_template_render() -> None:
    """Test rendering a template."""
    template = picofun.template.Template("tests/data/templates")
    assert template.render("empty.j2") == ""



"""Manage requirements.txt. file contents."""

import os
import re


class RequirementsManager:
    """Manage the requirements file."""

    def __init__(self, path: str) -> None:
        """Initialise the requirements manager."""
        self._path = path
        os.makedirs(os.path.dirname(self._path), exist_ok=True)
        self._requirements = self._load()

    def _load(self) -> list[str]:
        """Load the requirements file."""
        requirements = []
        if os.path.isfile(self._path):
            with open(self._path) as file:
                requirements = [line.strip() for line in file.readlines()]

        return requirements

    def save(self) -> None:
        """Save the requirements file."""
        with open(self._path, "w") as file:
            file.write("".join(self._requirements))

    def update(self, package: str, version: str) -> str:
        """Update the requirements file."""
        package = "picorun"
        # Subset of https://pip.pypa.io/en/stable/reference/requirement-specifiers/#requirement-specifiers
        regex = re.compile(rf"{package}\s?([~<>=]?=?|$)")

        found = False
        for line in self._requirements:
            if regex.match(line):
                found = True
                break

        if not found:
            self._requirements.append(f"{package}=={version}")
            self._requirements.sort()



"""Entry point for the application script."""

import logging

import picofun.cli
from picofun import __app_name__

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def main() -> None:
    """Run the application."""
    picofun.cli.app(prog_name=__app_name__)


if __name__ == "__main__":
    main()
